#Standard Python libary imports
import time
import joblib
import os
from sklearn.pipeline import Pipeline 
from sklearn.feature_selection import SelectKBest


#3rd party imports
from hyperopt import fmin, tpe, STATUS_OK, Trials

#Local imports
from mlautomator.objectives.classifier_objectives import Classifiers
from mlautomator.objectives.regressor_objectives import Regressors
from mlautomator.search_spaces import get_space, classifiers, regressors
from mlautomator.get_model import get_model


class MLAutomator:
    '''A tool for automating the algorithm tuning process in data science applications.

    MLAutomator leverages the intelligent search properties of Hyperopt to reduce 
    hyperparameter tuning times across large hyperparameter search spaces.  This extra 
    time allows you to spot-check a larger cross section of algorithms on your dataset.  

    MLAutomator does not perform predictions, its sole function is to find the optimal pre-processors, features, and 
    hyperparameters.

    Args:
        x_train (numpy ndarray): The training data that the models will be trained on.
        y_train (numpy ndarray): The target variables for the model.
        algo_type (str, optional, default='classifier'): Accepts 'classifier' or 'regressor'.  
        score_metric (str, optional, default='accuracy'): The scoring metric that Hyperopt will minimize on.  
        iterations (int, optional, default=25): The number of trials that Hyperopt will run on each algorithm candidate.
        num_cv_folds (int, optional, default=10): The number of folds to use in cross validation.
        repeats (int, optional, default=1): The number of passes to perform on cross validation.
        specific_algos (list of strings, default=None): A list of objective keys to overide comprehensive search.

    Public Attributes:
        start_time (time object): Used to measure elapsed time during training.
        count (int): A count of the total number of search space permutations analyzes.
        master_results (list): A history of all search spaces and their results.
        x_train (numpy ndarray): The training data passed to MLAutomator.
        y_train (numpy ndarray): The target data passed to MLAutomator.
        type (string): The type of model being built - classifier or regressor.
        score_metric (string): The score metric models are being optimized on.
        iterations (int): The number of iterations to perform on each objective function.
        num_cv_folds (int): The number of folds to use in cross validation.
        repeats (int): The number of repeats to perform in cross-validations.
        best_space (dict): The best search space discovered in the optimization process.
        best_algo (string): The description of the top performing algorithm.
        num_features (int): The number of features in the training data.
        num_samples (int): The number of samples in the training (and target) data.
    '''

    def __init__(
        self,
        x_train, 
        y_train,
        algo_type = 'classifier', 
        score_metric = 'accuracy',
        iterations = 25, 
        num_cv_folds = 10, 
        repeats = 1,
        specific_algos = None,
    ):
        # Beginning of initialization
        self.count = 0
        self.master_results = []
        self.x_train = x_train.copy()
        self.y_train = y_train.copy()
        self.type = algo_type
        self.score_metric = score_metric
        self.iterations = iterations
        self.num_cv_folds = num_cv_folds
        self.repeats = repeats
        self._initialize_best()
        self.num_features = self.x_train.shape[1]
        self.num_samples = self.x_train.shape[0]
        self.specific_algos = specific_algos

        # Initialize empty attributes
        self._objective = None
        self.keys = None
        self.best_space = None
        self.best_algo = None
        self.found_best_on = None
        self.best_pipeline = None
        self.start_time = None
        self.top_performer = None


    def __repr__(self):
        return f'MLAutomator ({self.type}, {self.score_metric}, Iterations: {self.iterations})'   


    def _initialize_best(self):
        '''Utility method for properly initializing the 'best' parameter.  
        
        After each iteration, the score will be checked
        against 'best' to see if the space used in that iteration was superior.  Depending on the scoring 
        metric used, 'best' needs to be initialized to different values.
        '''
        initializer_dict = {
            'accuracy': 0, 
            'neg_log_loss': 5,
            'neg_mean_squared_error' : 10000000,
        }
        self.best = initializer_dict[self.score_metric]


    def _get_objective(self,obj):
        '''A dictionary look-up utility that offers a clean way of looping through the objective 
        functions by key and returning the function call.  

        Args:
            obj (string): Key value representing the ojective function to call.

        Returns:
            <objective_list[obj]> (function call): A call to the appropriate objective function.    
        '''

        if self.type == 'classifier':
            objective_list = {        
                        '01': Classifiers.objective01,
                        '02': Classifiers.objective02,
                        '03': Classifiers.objective03,
                        '04': Classifiers.objective04,
                        '05': Classifiers.objective05,
                        '06': Classifiers.objective06,
                        '07': Classifiers.objective07,
                    }   
        else:  
            objective_list= {        
                        '01': Regressors.objective01,
                        '02': Regressors.objective02,
                        '03': Regressors.objective03,
                        '04': Regressors.objective04,
                        '05': Regressors.objective05,
                    }     


        return objective_list[obj]


    def _minimize_this(self,space):
        '''This is the "function to be minimized" by hyperopt. 
        
        This gets passed to the fmin function within the method find_best_algorithm.

        Args:
            space (dict): Subset of total search space selected by Hyperopt.
        '''
        iter_start = time.time()
        loss, algo = self._objective(self, space)
        self.count += 1

        # time methods for providing analytics on how each iteration is taking.
        iter_end = round((time.time()-iter_start)/60, 3)
        total_time = round((time.time()-self.start_time) / 60, 3)
        avg_time = round((total_time / self.count), 3)
        
        # If current loss is better than the best score reset best space, 
        # best score, and best algorithm.
        if loss < self.best:  
            self.best = loss
            self.best_space = space
            self.best_algo = algo
            self.found_best_on = self.count
            print('')
            print('new best score:', self.best)
            for key,values in space.items():
                print(str(key) + ' : ' + str(values))
            print('')    
            
        else:  
            str1 = 'Scanning '+algo+'.'
            str2 =' No Improvement. Iter time: '+str(iter_end)+'.'
            str3 =' Total Time Elapsed: '+str(total_time)+'.'
            str4 =' AVG Time: '+str(avg_time)
            print(str1 + str2 + str3 + str4) 

        self.master_results.append([loss, space])

        
        return {'loss': loss, 'status': STATUS_OK}


    def find_best_algorithm(self):   
        '''Finds the top permorning algorithm on the data.  
        
        It loops through each objective function that is 
        provided in the array 'objectives' and passes it to the Hyperopt function fmin.  fmin 
        will intelligently search the search spaces for each algorithm and attempt to minimize (optimize) 
        the scoring function provided.
        '''
        self.start_time = time.time()

        #If use provides specific objective list use that. 
        if self.specific_algos:
            objectives = self.specific_algos
        #Otheriwse, retrieve all objective keys from corresponding space.    
        else:    
            objectives = self.get_obj_key_list()

        #Pass each objective function to Hyperopt.
        for obj in objectives:
            self._objective = self._get_objective(obj)
            space = obj  
            trials = Trials()
            fmin(
                fn = self._minimize_this,
                space = get_space(self, space),
                algo = tpe.suggest,
                max_evals = self.iterations,
                trials = trials,
            )    


    def get_obj_key_list(self):
        '''Retrieves list of keys from classifier/regressor search spaces.
        '''
        if self.type == 'classifier':
            return classifiers().keys()
        elif self.type == 'regressor':
            return regressors().keys()   


        return []


    def print_best_space(self):

        if not self.best_space:
            print('Best space has not been determined yet. No models have tried on this data.')
            return 

        '''Prints out a report with the best algorithm and its configuration.
        '''
        print('Best Algorithm Configuration:')
        print('    ' + 'Best algorithm: '+ self.best_algo)
        print('    ' + 'Best ' + self.score_metric + ' : ' + str(self.best))
        for key,val in self.best_space.items():
            print('    '+ str(key)+' : '+ str(val), end='\n')                                
        print('    ' + 'Found best solution on iteration '+ str(self.found_best_on) + ' of ' + str(self.count)) 
        print('    ' + 'Validation used: ' +str(self.num_cv_folds) + '-fold cross-validation')   


    def fetch_best_performer(self):

        if not self.best_space:
            return None
        # Set top performer as dictionary object
        performance = {
            'Best Algorithm': self.best_algo, 
            'Score metric': self.score_metric,
            'Best score':str(self.best),
            'Validation used': str(self.num_cv_folds)+'-fold cross validation'
        }

        self.top_performer = {**performance, **self.best_space}
        
        return self.top_performer

    def save_best_pipeline(self, directory, filename='pipeline.joblib'):

        # In case the find_best_algorithm method hasn't been run yet...
        if not self.best_pipeline:
            print('No pipeline has been fit. Try running fit_best_pipeline() first.')
            return

        if not directory: 
            print ('No directory has been specified and is a requirement.')
            return    

        save_as = os.path.join(directory, filename)
        # Export the pipeline to a file
        joblib.dump(self.best_pipeline, open(save_as, 'wb'))
        print('Best model has been saved to: '+save_as)


    def load_best_pipeline(self, filename):

        if not filename: 
            print ('No directory has been specified and is a requirement.')
            return    

        # Load the pipeline from a file
        self.best_pipeline = joblib.load(filename)     
 

    def fit_best_pipeline(self):

        # In case the find_best_algorithm method hasn't been run yet...
        if not self.best_space:
            print('Best model has not been determined. Run find_best_algorithm() first.')
            return

        print('\n')
        print('Fitting best pipeline...')    

        # Assemble pipeline variables
        k_best = self.best_space['k_best']
        scaler = self.best_space['scaler']
        model = get_model(self.best_algo, self.best_space)

        # Build optimal pipeline again
        pipeline = []
        pipeline = Pipeline([
            ('scaler', scaler),
            ('select_best', SelectKBest(k = k_best)),
            ('classifier', model),
        ])

        print(model.get_params)
        self.best_pipeline = pipeline.fit(self.x_train, self.y_train)
        print('Completed fitting the model on entire dataset with optimal parameters')
     
    



#Standard Python Libary imports
from sklearn.datasets import (
    load_boston,
    load_iris,
    load_diabetes,
    load_digits,
    load_linnerud,
    load_wine,
    load_breast_cancer,
)

def get_data(key):
    '''
    Accepts a text string as a key and calls the appropriate data loading function from 
    sklearn.  

    Args:
        key (str) : The description of the dataset to load. 

    Returns:
        library[key] : functional call to the appropriate dataset from sklearn.    
    '''
    library = {
        'boston': load_boston(),
        'iris' : load_iris(),
        'diabetes' : load_diabetes(), 
        'digits' : load_digits(), 
        'linnerud' : load_linnerud(),
        'wine' : load_wine(), 
        'breast_cancer' : load_breast_cancer(),
    }

    return library[key]



#Classifier imports
from sklearn.neighbors import KNeighborsClassifier
from sklearn.linear_model import SGDClassifier
from sklearn.ensemble import RandomForestClassifier
from sklearn.naive_bayes import GaussianNB
from sklearn.decomposition import PCA
from sklearn.preprocessing import StandardScaler, RobustScaler
from sklearn.svm import SVC
from sklearn.ensemble import BaggingClassifier
from sklearn.discriminant_analysis import LinearDiscriminantAnalysis
from sklearn.linear_model import LogisticRegression
from xgboost import XGBClassifier

#regressor imports
from sklearn.neighbors import KNeighborsRegressor
from sklearn.linear_model import SGDRegressor, LogisticRegression
from sklearn.ensemble import RandomForestRegressor, BaggingRegressor
from sklearn.naive_bayes import GaussianNB
from xgboost import XGBRegressor
from sklearn.svm import SVR
from sklearn.discriminant_analysis import LinearDiscriminantAnalysis

#Local imports
from mlautomator.search_keys import get_keys




def get_model(algo, space_dict):

    keys = get_keys(algo)
    space = {k:space_dict[k] for k in set(space_dict).intersection(keys)}

    model_lib = { 
        'xgboost_classifier': XGBClassifier,
        'xgboost_regressor':  XGBRegressor,
        'SGDClassifier': SGDClassifier,
        'SGDRegressor': SGDRegressor,
        'RandomForestClassifier': RandomForestClassifier,
        'RandomForestRegressor': RandomForestRegressor,
        'SVC': SVC,
        'SVR': SVR,          
        'LogisticRegression' : LogisticRegression,
        'KNeighborClassifier': KNeighborsClassifier,
        'KNeighborRegressor': KNeighborsRegressor,
        'GaussianNB': GaussianNB,
    }

    return model_lib[algo](**space)


#Standard Python libary imports
import numpy as np
from sklearn.preprocessing import StandardScaler, RobustScaler, MinMaxScaler, PowerTransformer

#3rd party imports
from hyperopt import hp

SCALER_LIST = [
        StandardScaler(copy=True), 
        RobustScaler(copy=True), 
        MinMaxScaler(copy=True), 
        PowerTransformer(), 
        None
        ]

N_COMPONENTS = np.round(np.arange(0.1, 0.99, 0.01), 5)        


def classifiers(**kwargs):

    # If there variable parameters to set they get passed through kwargs.
    k_best=kwargs.get('k_best', [3])

    CLASSIFIER_SPACES = {
                #XGBoost Classifier search space.
                '01':      {
                            'max_depth': hp.choice('x_max_depth', [3, 4, 5, 6]),
                            'min_child_weight': hp.choice('x_min_child_weight', np.round(np.arange(0.0001, 0.3, 0.0001), 5)),
                            'gamma': hp.choice('x_gamma', np.round(np.arange(0.0,40.0,0.005),5)),
                            'learning_rate': hp.choice('x_learning_rate',np.round(np.arange(0.0005,0.3,0.0005),5)),
                            'subsample': hp.choice('x_subsample',np.round(np.arange(0.01,1.0,0.01),5)),
                            'colsample_bylevel': hp.choice('x_colsample_bylevel', np.round(np.arange(0.1,1.0,0.01),5)),
                            'colsample_bytree': hp.choice('x_colsample_bytree', np.round(np.arange(0.1,1.0,0.01),5)),
                            'n_estimators': hp.choice('x_n_estimators', np.arange(25,850,1)),
                            'k_best': hp.choice('x_k_best', k_best),
                            'scaler': hp.choice('x_scale', SCALER_LIST),
                            },        
            
                #SGD Classifier search space    
                '02':       {
                            'loss': hp.choice('x_loss', ['log','modified_huber']),
                            'penalty': hp.choice('x_penalty', ['none', 'l2', 'l1', 'elasticnet']),
                            'alpha': hp.choice('x_alpha', np.arange(0.0,0.5,0.00001)),
                            'max_iter': hp.choice('x_max_iter', np.arange(5,1000,1)),
                            'k_best': hp.choice('x_k_best', k_best),
                            'n_components': hp.choice('x_n_comps', N_COMPONENTS),
                            'scaler': hp.choice('x_scale', SCALER_LIST),
                            },
                        
                #Random Forest search space        
                '03':       {
                            'n_estimators': hp.choice('x_loss', np.arange(50,750,5)),
                            'max_depth': hp.choice('x_max_depth', np.arange(1,8,1)),
                            'max_features': hp.choice('x_max_features', ['sqrt','log2',None]),
                            'criterion': hp.choice('x_crit', ['gini','entropy']),
                            'min_samples_split': hp.choice('x_mss', np.arange(0.000001,0.3,0.00005)),
                            'min_samples_leaf': hp.choice('x_msl', np.arange(0.000001,0.3,0.00005)),
                            'min_impurity_decrease': hp.choice('min_id', np.arange(0.000001,0.3,0.00005)),
                            'n_jobs': hp.choice('x_njobs',[-1]),
                            'k_best': hp.choice('x_k_best',k_best),
                            'n_components': hp.choice('x_n_comps', N_COMPONENTS),
                            'scaler': hp.choice('x_scale', SCALER_LIST)                    
                            },
                        
                #Support Vector Machine search space        
                '04':       {
                            'C': hp.choice('C', np.arange(0.0,1.0,0.00005)),
                            'gamma': hp.choice('x_gamma', ['auto']),
                            'kernel': hp.choice('x_kernel', ['poly', 'rbf']),
                            'degree': hp.choice('x_degree', [2,3,4,5]),
                            'probability': hp.choice('x_probability', [True]),
                            'n_estimators': hp.choice('x_n_estimators', np.arange(4, 25, 1)),
                            'k_best': hp.choice('x_k_best', k_best),
                            'scaler': hp.choice('x_scale', SCALER_LIST)                    
                            },

                #GaussianNB   
                '05':       {
                            'k_best': hp.choice('x_k_best', k_best),
                            'scaler': hp.choice('x_scale', SCALER_LIST)                    
                            },      

                #LogisticRegression
                '06':       {
                            'penalty': hp.choice('x_penalty', ['l1','l2']),
                            'C': hp.choice('C', np.round(np.arange(0.0, 1.0, 0.00001), 5)),
                            'solver': hp.choice('x_solver', [ 'lbfgs', 'liblinear', 'sag', 'saga']),
                            'k_best': hp.choice('x_k_best', k_best),
                            'scaler': hp.choice('x_scale', SCALER_LIST)                                
                            },

                #KNearestNeighbor
                '07':       {
                            'n_neighbors': hp.choice('x_n_neighbors', np.arange(1,10,1)),
                            'weights': hp.choice('x_weights', ['uniform','distance']),
                            'algorithm': hp.choice('x_algorithm', ['ball_tree','kd_tree','brute','auto']),  
                            'k_best': hp.choice('x_k_best', k_best),
                            'scaler': hp.choice('x_scale', SCALER_LIST),
                            },   
                        
                }


    return CLASSIFIER_SPACES        


def regressors(**kwargs):
    '''
    A method for return a dictionary of parameters from a selected search space. 
    kwargs is passed because some search spaces have a variable parameter space that is data and
    algorithm dependent.  
    '''

    # If there variable parameters to set they get passed through kwargs.
    k_best=kwargs.get('k_best', [3])

    REGRESSION_SPACES={
                #XGBoost Regressor search space.
                '01':      {
                            'max_depth': hp.choice('x_max_depth', [3,4,5,6]),
                            'min_child_weight': hp.choice('x_min_child_weight', np.round(np.arange(0.0001, 0.3, 0.0001), 5)),
                            'gamma': hp.choice('x_gamma', np.round(np.arange(0.0,40.0,0.005), 5)),
                            'learning_rate': hp.choice('x_learning_rate', np.round(np.arange(0.0005,0.3,0.0005), 5)),
                            'subsample': hp.choice('x_subsample', np.round(np.arange(0.01,1.0,0.01),5)),
                            'colsample_bylevel': hp.choice('x_colsample_bylevel', np.round(np.arange(0.1,1.0,0.01), 5)),
                            'colsample_bytree': hp.choice('x_colsample_bytree', np.round(np.arange(0.1,1.0,0.01), 5)),
                            'n_estimators': hp.choice('x_n_estimators', np.arange(25,850,1)),
                            'k_best': hp.choice('x_k_best', k_best),
                            'n_components': hp.choice('x_n_comps', N_COMPONENTS),
                            'scaler': hp.choice('x_scale', SCALER_LIST)
                            },        
            
                #SGD Regressor search space    
                '02':       {
                            'loss': hp.choice('x_loss', ['squared_loss','huber']),
                            'penalty': hp.choice('x_penalty', ['none', 'l2', 'l1', 'elasticnet']),
                            'alpha': hp.choice('x_alpha', np.arange(0.0,0.5,0.00001)),
                            'max_iter': hp.choice('x_max_iter', np.arange(5,1000,1)),
                            'k_best': hp.choice('x_k_best', k_best),
                            'n_components': hp.choice('x_n_comps', N_COMPONENTS),
                            'scaler': hp.choice('x_scale', SCALER_LIST)
                            },
                        
                #Random Forest Regressor search space        
                '03':       {
                            'n_estimators': hp.choice('x_loss', np.arange(50, 750, 5)),
                            'max_depth': hp.choice('x_max_depth', np.arange(1, 8, 1)),
                            'max_features': hp.choice('x_max_features', ['sqrt','log2',None]),
                            'criterion': hp.choice('x_crit', ['mse']),
                            'min_samples_split': hp.choice('x_mss', np.arange(0.000001, 0.3, 0.00005)),
                            'min_samples_leaf': hp.choice('x_msl', np.arange(0.000001, 0.3, 0.00005)),
                            'min_impurity_decrease': hp.choice('min_id', np.arange(0.000001, 0.3, 0.00005)),
                            'n_jobs': hp.choice('x_njobs', [-1]),
                            'k_best': hp.choice('x_k_best', k_best),
                            'n_components': hp.choice('x_n_comps', N_COMPONENTS),
                            'scaler': hp.choice('x_scale', SCALER_LIST)                    
                            },
                        
                #Support Vector Machine Regressor search space        
                '04':       {
                            'C': hp.choice('C', np.arange(0.0, 1.0, 0.00005)),
                            'gamma': hp.choice('x_gamma', ['auto']),
                            'kernel': hp.choice('x_kernel', ['poly', 'rbf']),
                            'degree': hp.choice('x_degree', [2, 3, 4]),
                            'n_estimators': hp.choice('x_n_estimators', np.arange(4, 25, 1)),
                            'k_best': hp.choice('x_k_best', k_best),
                            'scaler': hp.choice('x_scale', SCALER_LIST)                    
                            },

                #KNearestNeighbor Regressor search space
                '05':       {
                            'n_neighbors': hp.choice('x_n_neighbors', np.arange(1,10,1)),
                            'weights': hp.choice('x_weights', ['uniform','distance']),
                            'algorithm': hp.choice('x_algorithm', ['ball_tree','kd_tree','brute','auto']),  
                            'k_best': hp.choice('x_k_best', k_best),
                            'scaler': hp.choice('x_scale', SCALER_LIST),
                            },                      
                }


    return REGRESSION_SPACES            


def get_space(automator, space):
    '''
    Get space checks the algo_type property in the automator object and returns the search space that corresponds 
    with the provided key. 

    Args:
        automator (object): an MLAutomator object.
        space (str): The key value for the search space to return.   

    Returns:    
        <search space> (dict) :A selection of hyperparameters chosen for analysis in the optimization engine.   
    '''
    
    # include any parameters that are data-dependent here
    algo_dependent_params = {
        'k_best' : np.arange(2, automator.num_features-1, 1),
        }
 
    if automator.type == 'classifier':
        Space_List = classifiers(**algo_dependent_params)

    elif automator.type == 'regressor':     
        Space_List = regressors(**algo_dependent_params)


    return Space_List[space]



#Standard python library imports
import unittest
import os

#Local imports
from mlautomator.mlautomator import MLAutomator
from mlautomator.search_spaces import classifiers, regressors, get_space
from mlautomator.search_keys import get_keys, ALGORITHM_KEYS
from data.utilities import clf_prep
from tempfile import mkdtemp



class TestMLAutomator(unittest.TestCase):

    directory=mkdtemp()
    x, y = clf_prep('pima-indians-diabetes.csv')
    automator=MLAutomator(x,y, iterations = 2, specific_algos=['01'])
    automator.find_best_algorithm()

    def test_fit_best_pipeline(self):
        self.automator.fit_best_pipeline()
        self.assertIsNotNone(self.automator.best_pipeline)
        print(self.automator.best_pipeline)

    def test_model_dump(self):
        self.automator.save_best_pipeline(self.directory)    

    def test_model_load(self):
        self.automator.load_best_pipeline(filename=self.directory+'/pipeline.joblib')           


#Standard python library imports
import unittest

#Local imports
from mlautomator.mlautomator import MLAutomator
from mlautomator.search_spaces import classifiers, regressors, get_space
from mlautomator.search_keys import get_keys, ALGORITHM_KEYS
from data.utilities import clf_prep


class TestMLAutomator(unittest.TestCase):

    x, y = clf_prep('pima-indians-diabetes.csv')
    automator=MLAutomator(x,y)

    def test_automator_initialization(self):
        '''
        Test that all class properties are being initialized properly.
        '''
        
        self.assertEqual(self.automator.best,0)
        self.assertEqual(self.automator.count,0)
        self.assertEqual(self.automator.start_time,None)
        self.assertEqual(self.automator._objective,None)
        self.assertEqual(self.automator.keys,None)
        self.assertEqual(self.automator.master_results,[])
        self.assertEqual(self.automator.type,'classifier')
        self.assertEqual(self.automator.score_metric,'accuracy')
        self.assertEqual(self.automator.iterations,25)
        self.assertEqual(self.automator.num_cv_folds,10)
        self.assertEqual(self.automator.repeats,1)

    def test_get_obj_key_list(self):
        self.assertIsNotNone(classifiers().keys())
        self.assertIsNotNone(regressors().keys())

    def test_get_keys(self):
        for key in ALGORITHM_KEYS.keys():
            self.assertIsNotNone(get_keys(key))
            print(get_keys(key))    

    def test_get_space_regressors(self):
        for key in regressors().keys():
            self.assertIsNotNone(get_space(self.automator, key))        

    def test_get_space_classifiers(self):
        for key in classifiers().keys():
            self.assertIsNotNone(get_space(self.automator, key))

    def test_user_feedback_when_best_space_not_evaluated(self):
        self.assertIsNone(self.automator.print_best_space()) 
        self.assertIsNone(self.automator.save_best_pipeline(directory = None)) 
        self.assertIsNone(self.automator.fit_best_pipeline())                                      


# Standard Python Library imports
import time
import numpy as np
import pandas as pd
from sklearn.model_selection import (StratifiedKFold, RepeatedKFold, KFold, cross_val_score)
from sklearn.neighbors import KNeighborsClassifier
from sklearn.linear_model import SGDClassifier
from sklearn.ensemble import RandomForestClassifier
from sklearn.naive_bayes import GaussianNB
from sklearn.decomposition import PCA
from sklearn.preprocessing import StandardScaler, RobustScaler
from sklearn.svm import SVC
from sklearn.ensemble import BaggingClassifier
from sklearn.discriminant_analysis import LinearDiscriminantAnalysis
from sklearn.linear_model import LogisticRegression
from sklearn.pipeline import Pipeline, FeatureUnion
from sklearn.feature_selection import SelectKBest, chi2,f_classif
import warnings

#Local Imports
from mlautomator.search_keys import get_keys

#3rd party imports
from xgboost import XGBClassifier

warnings.filterwarnings("ignore")

class Classifiers:
    '''
    A utility class for holding all the objective functions for each classifier in the search space.
        - You will see commonalities amongst the objective functions.  
        - Each objective function takes an MLautomator object and "parameter space" as arguments.  
        - During each pass through this objective function, Hyperopt selects a subset of the search space for 
          the appropriate algorithm in search_spaces.py.
        - Note that Hyperopt is also calling permutations of data transforms and feature selection as well.
        - Each objective function returns the mean cross-validated score, and the name of the algorithm.  
        - This gets anaylzed and packaged in the automator class itself.
    '''

    @staticmethod
    def objective01(automator, space):
        '''
        Objective function for XGBoost Classifier.
        '''
        algo = 'xgboost_classifier'
        X = automator.x_train
        Y = automator.y_train
        
        #Define the subset of dictionary keys that should get passed to the machine learning
        #algorithm.
        
        keys = get_keys(algo)
        subspace = {k:space[k] for k in set(space).intersection(keys)}
        
        #Extract the remaining keys that are pertinent to data preprocessing.
        
        model = XGBClassifier(n_jobs=-1, **subspace)     
        scaler = space.get('scaler')
        num_features = space.get('k_best')
        
        #Assemble a data pipeline with the extracted data preprocessing keys.
        pipeline = []
        pipeline = Pipeline([
            ('scaler', scaler),
            ('select_best', SelectKBest(k = num_features)),
            ('classifier', model),
        ])
        
        #perform cross validation and return the mean score.
        kfold = RepeatedKFold(n_splits = automator.num_cv_folds, n_repeats = automator.repeats)

        try:
            scores = -cross_val_score(pipeline, X, Y, cv=kfold, scoring = automator.score_metric, verbose=False).mean()  
        except ValueError:
            print('An error occurred with the following space: ')
            print(space)
            return automator.best, algo    

        return scores, algo


    @staticmethod
    def objective02(automator, space):
        '''
        Objective function for SGD Classifier.
        '''
        algo = 'SGDClassifier'
        X = automator.x_train
        Y = automator.y_train
        #Define the subset of dictionary keys that should get passed to the machine learning
        #algorithm.
        
        keys = get_keys(algo)   
        subspace = {k:space[k] for k in set(space).intersection(keys)}
        
        #Extract the remaining keys that are pertinent to data preprocessing.
        model = SGDClassifier(n_jobs=-1, **subspace)   
        scaler = space.get('scaler')
        num_features = space.get('k_best')
        
        #Assemble a data pipeline with the extracted data preprocessing keys.
        pipeline = []
        pipeline = Pipeline([
            ('scaler', scaler),
            ('select_best', SelectKBest(k=num_features)),
            ('classifier', model),
        ])
        
        #perform cross validation and return the mean score.
        kfold = RepeatedKFold(n_splits=automator.num_cv_folds, n_repeats=automator.repeats)

        try:
            scores = -cross_val_score(pipeline, X, Y, cv=kfold, scoring=automator.score_metric, verbose=False).mean()  
        except ValueError:
            print('An error occurred with the following space: ')
            print(space)
            return automator.best, algo       

        return scores, algo


    @staticmethod
    def objective03(automator,space):
        '''
        Objective function for Random Forest Classifier.
        '''
        algo = 'RandomForestClassifier'
        X = automator.x_train
        Y = automator.y_train
        #Define the subset of dictionary keys that should get passed to the machine learning
        #algorithm.
        
        keys = get_keys(algo)  
        subspace = {k:space[k] for k in set(space).intersection(keys)}
        
        #Extract the remaining keys that are pertinent to data preprocessing.
        model = RandomForestClassifier(**subspace)   
        scaler = space.get('scaler')
        num_features = space.get('k_best')
        
        #Assemble a data pipeline with the extracted data preprocessing keys.
        pipeline = []
        pipeline = Pipeline([
            ('scaler', scaler),
            ('select_best', SelectKBest(k=num_features)),
            ('classifier',model),
        ])
        
        #perform two passes of 10-fold cross validation and return the mean score.
        kfold = RepeatedKFold(n_splits=10, n_repeats=1)

        try:
            scores = -cross_val_score(pipeline, X, Y, cv=kfold, scoring=automator.score_metric, verbose=False).mean()  
        except ValueError:
            print('An error occurred with the following space: ')
            print(space)
            return automator.best, algo    

        return scores, algo


    @staticmethod
    def objective04(automator,space):
        '''
        Objective function for Support Vector Machines. Note that this method uses a Bagged Classifier 
        as a wrapper for SVC.  Support Vector Machine run time scales by O(N^3).  Using bagged classifiers
        break up the dataset into smaller samples so that runtime is manageable.
        '''
        algo = 'SVC'
        X = automator.x_train
        Y = automator.y_train

        #Define the subset of dictionary keys that should get passed to the machine learning
        #algorithm.
        
        keys = get_keys(algo)    
        subspace = {k:space[k] for k in set(space).intersection(keys)}
 
        #Build a model with the parameters from our Hyperopt search space.

        n_estimators = space.get('n_estimators')
        model = BaggingClassifier(
            SVC(probability=True, **subspace),
            max_samples=automator.num_samples // n_estimators,
            n_estimators = n_estimators,
            n_jobs = -1)   

        scaler = space.get('scaler')
        num_features = space.get('k_best')
        
        #Assemble a data pipeline with the extracted data preprocessing keys.
        pipeline = []
        pipeline = Pipeline([
            ('scaler', scaler),
            ('select_best', SelectKBest(k = num_features)),
            ('classifier', model),
        ])
        
        #perform cross validation and return the mean score.
        kfold = RepeatedKFold(n_splits=automator.num_cv_folds, n_repeats=automator.repeats)

        try:
            scores = -cross_val_score(pipeline, X, Y, cv=kfold, scoring=automator.score_metric, verbose=False).mean()  
        except ValueError:
            print('An error occurred with the following space: ')
            print(space)
            return automator.best, algo     

        return scores, algo


    @staticmethod
    def objective05(automator,space):
        '''
        Objective function for Naive Bayes
        '''
        algo = 'GaussianNB'
        X = automator.x_train
        Y = automator.y_train

        #Build a model with the parameters from our Hyperopt search space.
        model = GaussianNB()
        scaler = space.get('scaler')
        num_features = space.get('k_best')
        
        #Assemble a data pipeline with the extracted data preprocessing keys.
        pipeline = []
        pipeline = Pipeline([
            ('scaler', scaler),
            ('select_best', SelectKBest(k=num_features)),
            ('classifier', model),
        ])
        
        #perform cross validation and return the mean score.
        kfold = RepeatedKFold(n_splits=automator.num_cv_folds, n_repeats=automator.repeats)

        try:
            scores = -cross_val_score(pipeline, X, Y, cv=kfold, scoring=automator.score_metric, verbose=False).mean()  
        except ValueError:
            print('An error occurred with the following space: ')
            print(space)
            return automator.best, algo    

        return scores, algo


    @staticmethod
    def objective06(automator,space):
        '''
        Objective function for Logistic Regression.
        '''
        algo = 'LogisticRegression'
        X = automator.x_train
        Y = automator.y_train

        #Define the subset of dictionary keys that should get passed to the machine learning
        #algorithm.
        keys = get_keys(algo)    
        subspace = {k:space[k] for k in set(space).intersection(keys)}      

        #Build a model with the parameters from our Hyperopt search space.
        model = LogisticRegression(**subspace)
        scaler=space.get('scaler')
        num_features=space.get('k_best')
        
        #Assemble a data pipeline with the extracted data preprocessing keys.
        pipeline=[]
        pipeline=Pipeline([
            ('scaler', scaler),
            ('select_best', SelectKBest(k=num_features)),
            ('classifier',model),
        ])
        
        #perform cross validation and return the mean score.
        kfold = RepeatedKFold(n_splits=automator.num_cv_folds, n_repeats=automator.repeats)

        try:
            scores = -cross_val_score(pipeline, X, Y, cv=kfold, scoring=automator.score_metric, verbose=False).mean()  
        except ValueError:
            print('An error occurred with the following space: ')
            print(space)
            return automator.best, algo    

        return scores, algo        


    @staticmethod
    def objective07(automator,space):
        '''
        Objective function for K-Nearest Neighbors Voting Classifier.
        '''
        algo = 'KNeighborClassifier'
        X = automator.x_train
        Y = automator.y_train

        #Define the subset of dictionary keys that should get passed to the machine learning
        #algorithm.
        keys = get_keys(algo)    
        subspace = {k:space[k] for k in set(space).intersection(keys)}      

        #Build a model with the parameters from our Hyperopt search space.
        model = KNeighborsClassifier(n_jobs=-1, **subspace)
        scaler = space.get('scaler')
        num_features = space.get('k_best')
        
        #Assemble a data pipeline with the extracted data preprocessing keys.
        pipeline = []
        pipeline = Pipeline([
            ('scaler', scaler),
            ('select_best', SelectKBest(k=num_features)),
            ('classifier',model),
        ])
        
        #perform cross validation and return the mean score.
        kfold = RepeatedKFold(n_splits=automator.num_cv_folds, n_repeats=automator.repeats)

        try:
            scores = -cross_val_score(pipeline, X, Y, cv=kfold, scoring=automator.score_metric, verbose=False).mean()  
        except ValueError:
            print('An error occurred with the following space: ')
            print(space)
            return automator.best, algo    
            
        return scores, algo   


#Standard python imports
import unittest
import warnings

#Local imports
from mlautomator.mlautomator import MLAutomator
from data.utilities import from_sklearn
from mlautomator.objectives import *

warnings.filterwarnings("ignore")

class TestRegressors(unittest.TestCase):
    '''
    Integrations tests for regressor objective functions. Checks to see that at least 
    one iteration can be executed on the given dataset and algorithm.

    To increase test speed:
        - iterations set to 1
        - num_cv_folds=2
    '''

    iters=1
    folds=2
    x, y = from_sklearn('boston')
    print('Data prepped')


    def test_objective01_xgboost_regressor(self):
        automator=MLAutomator(
            self.x, 
            self.y, 
            iterations=self.iters, 
            algo_type='regressor', 
            specific_algos=['01'], 
            num_cv_folds=self.folds, 
            score_metric = 'neg_mean_squared_error',
            )
        automator.find_best_algorithm()
        self.assertEqual(automator.best_algo, 'xgboost_regressor')

    def test_objective02_sgd_regressor(self):
        automator=MLAutomator(
            self.x, 
            self.y, 
            iterations=self.iters, 
            algo_type='regressor', 
            specific_algos=['02'], 
            num_cv_folds=self.folds, 
            score_metric = 'neg_mean_squared_error',
            )        
        automator.find_best_algorithm()
        self.assertEqual(automator.best_algo, 'SGDRegressor')

    def test_objective03_sgd_regressor(self):
        automator=MLAutomator(
            self.x, 
            self.y, 
            iterations=self.iters, 
            algo_type='regressor', 
            specific_algos=['03'], 
            num_cv_folds=self.folds, 
            score_metric = 'neg_mean_squared_error',
            )        
        automator.find_best_algorithm()
        self.assertEqual(automator.best_algo, 'RandomForestRegressor')

    def test_objective04_svr_regressor(self):
        automator=MLAutomator(
            self.x, 
            self.y, 
            iterations=self.iters, 
            algo_type='regressor', 
            specific_algos=['04'], 
            num_cv_folds=self.folds, 
            score_metric = 'neg_mean_squared_error',
            )        
        automator.find_best_algorithm()
        self.assertEqual(automator.best_algo, 'SVR')                

    def test_objective05_knn_regressor(self):
        automator=MLAutomator(
            self.x, 
            self.y, 
            iterations=self.iters, 
            algo_type='regressor', 
            specific_algos=['05'], 
            num_cv_folds=self.folds, 
            score_metric = 'neg_mean_squared_error',
            )        
        automator.find_best_algorithm()
        self.assertEqual(automator.best_algo, 'KNeighborRegressor')          



        


#Standard python imports
import unittest

#Local imports
from mlautomator.mlautomator import MLAutomator
from data.utilities import clf_prep
from mlautomator.objectives import *

class TestClassifiers(unittest.TestCase):
    '''
    Integrations tests for classifier objective functions. Checks to see that at least 
    one iteration can be executed on the given dataset and algorithm.

    To increase test speed:
        - iterations set to 1
        - num_cv_folds=2
    '''

    iters=1
    folds=2
    x, y = clf_prep('pima-indians-diabetes.csv')
    print('Data prepped')

    def test_objective01_xgboost(self):
        automator=MLAutomator(self.x, self.y, iterations=self.iters, specific_algos=['01'], num_cv_folds=self.folds)
        automator.find_best_algorithm()
        self.assertEqual(automator.best_algo, 'xgboost_classifier')

    def test_objective02_sgd_classifier(self):
        automator=MLAutomator(self.x, self.y, iterations=self.iters, specific_algos=['02'], num_cv_folds=self.folds)
        automator.find_best_algorithm()
        self.assertEqual(automator.best_algo, 'SGDClassifier')

    def test_objective03_sgd_classifier(self):
        automator=MLAutomator(self.x, self.y, iterations=self.iters, specific_algos=['03'], num_cv_folds=self.folds)
        automator.find_best_algorithm()
        self.assertEqual(automator.best_algo, 'RandomForestClassifier')

    def test_objective04_bag_of_svc(self):
        automator=MLAutomator(self.x, self.y, iterations=self.iters, specific_algos=['04'], num_cv_folds=self.folds)
        automator.find_best_algorithm()
        self.assertEqual(automator.best_algo, 'SVC')                

    def test_objective05_naive_bayes(self):
        automator=MLAutomator(self.x, self.y, iterations=self.iters, specific_algos=['05'], num_cv_folds=self.folds)
        automator.find_best_algorithm()
        self.assertEqual(automator.best_algo, 'GaussianNB')          

    def test_objective06_logistic_regression(self):
        automator=MLAutomator(self.x, self.y, iterations=self.iters, specific_algos=['06'], num_cv_folds=self.folds)
        automator.find_best_algorithm()
        self.assertEqual(automator.best_algo, 'LogisticRegression')    

    def test_objective07_knn(self):
        automator=MLAutomator(self.x, self.y, iterations=self.iters, specific_algos=['07'], num_cv_folds=self.folds)
        automator.find_best_algorithm()
        self.assertEqual(automator.best_algo, 'KNeighborClassifier')                    


  



#Standard Python libary imports
import pandas as pd

#Local imports
from data.data_dict import get_data


def clf_prep(filename=None, target_columns=1, ignore_columns=None):
    '''A utility method for preparing training and target data from a csv file.

    Args:
        filename (string):  the file name of the dataset to be imported.  This function assumes the
            dataset resides in 'data/datasets/'
        target_columns (int, optional, default=1): The number of target columns there (in case it 
            might be one-hot encoded, for example). Method assumes that target columns reside at 
            the end of the dataframe. The default value takes the last column as the target.
        ignore_columns (list, optional, default=None): A list of column names or indices that 
            you want dropped or ignored during preparation. These columns will not be included in the output.

    Returns:
        x_train (numpy ndarray): Training data for the model.
        y_train (numpy ndarray): Target data for the model.   
    '''

    #Load data from csv file.  
    data_directory = 'data/datasets/'
    training_data = pd.read_csv(data_directory+filename, header=0)

    if ignore_columns:
        training_data = training_data.drop(ignore_columns, axis=1)
    
    # Transform the loaded CSV data into numpy arrays
    columns = list(training_data)
    features = columns[:-1]
    feature_data = training_data[features]
    target_data = training_data[columns[-target_columns:]]
         
    x_train = feature_data.values
    y_train = target_data.values.ravel()


    return x_train, y_train


def from_sklearn(key):
    '''Imports and processes data from sklearn's library of sample datasets into 
    a format digestible by MLAutomator.

    Args:
        key (string): The key descripion of the dataset to pull from sklearn. See data_dict module for info
            on available datasets and their keys.

    Returns:
        x_train (numpy ndarray): Training data for the model.
        y_train (numpy ndarray): Target data for the model.   
    '''
    data = get_data(key)
    dataframe = pd.DataFrame(data.data)
    dataframe.columns = data.feature_names
    feature_data = dataframe[data.feature_names]
    x_train = feature_data.values
    y_train = data.target

    
    return x_train, y_train



'''A module for retrieving a list of all optimization parameters unique to each algorithm.
'''

ALGORITHM_KEYS = {
    'xgboost_classifier': (
        'max_depth',
        'min_child_weight',
        'gamma',
        'learning_rate',
        'subsample',
        'colsample_bylevel', 
        'colsample_bytree',
        'n_estimators',
        ),
    'xgboost_regressor': (
        'max_depth',
        'min_child_weight',
        'gamma',
        'learning_rate',
        'subsample',
        'colsample_bylevel', 
        'colsample_bytree',
        'n_estimators',
        ),            
    'SGDClassifier': (
        'loss',
        'penalty',
        'alpha',
        'max_iter',
        ),
    'SGDRegressor': (
        'loss',
        'penalty',
        'alpha',
        'max_iter',
        ),            
    'RandomForestClassifier': (
        'n_estimators',
        'max_depth',
        'max_features',
        'criterion',
        'min_samples_split',
        'min_samples_leaf',
        'min_impurity_decrease',
        'n_jobs',
        ),
    'RandomForestRegressor': (
        'n_estimators',
        'max_depth',
        'max_features',
        'criterion',
        'min_samples_split',
        'min_samples_leaf',
        'min_impurity_decrease',
        'n_jobs',
        ),            
    'SVC': (
        'C',
        'gamma',
        'kernel',
        'degree',
        ),
    'SVR': (
        'C',
        'gamma',
        'kernel',
        'degree',
        ),            
    'LogisticRegression' : (
        'C',
        'solver',
        'pentalty',   
        ),
    'KNeighborClassifier': (
        'n_neighbors',
        'weights',
        'algorithm',
        ),
    'KNeighborRegressor': (
        'n_neighbors',
        'weights',
        'algorithm',
        ),    
    'GaussianNB': (),                            
}


def get_keys(algorithm):
    '''
    Returns the dictionary of comprehensive search key parameters that Hyperopt will attempt
    to optimize on.  

    Args:
        algorithm (string): The key for the appropriate algorithm search parameters. 
            (i.e, 'xgboost_classifier')

    Returns:
        ALGORITHM_KEYS[algorithm] (dict): dictionary of search parameter keys.

    '''

    
    return ALGORITHM_KEYS[algorithm]


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()


setuptools.setup(
    name="mlautomator",
    version="1.1.3",
    author="Kevin Vecmanis",
    author_email="kevinvecmanis@gmail.com",
    description="A fast, simple way to train machine learning algorithms",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/VanAurum/ML-automator",
    packages=setuptools.find_packages(),
    classifiers=[
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        'numpy',
        'pandas',
        'sklearn',
        'hyperopt',
        'xgboost',
        'joblib',
    ]
)



# Standard Python Library imports
import time
import numpy as np
import pandas as pd
from sklearn.neighbors import KNeighborsRegressor
from sklearn.model_selection import (StratifiedKFold, RepeatedKFold, KFold, cross_val_score)
from sklearn.linear_model import SGDRegressor, LogisticRegression
from sklearn.ensemble import RandomForestRegressor, BaggingRegressor
from sklearn.naive_bayes import GaussianNB
from sklearn.decomposition import PCA
from sklearn.preprocessing import StandardScaler, RobustScaler
from sklearn.svm import SVR
from sklearn.discriminant_analysis import LinearDiscriminantAnalysis
from sklearn.pipeline import Pipeline, FeatureUnion
from sklearn.feature_selection import SelectKBest, chi2,f_classif
import warnings

#Local Imports
from mlautomator.search_keys import get_keys

#3rd party imports
from xgboost import XGBRegressor

warnings.filterwarnings("ignore")

class Regressors:
    '''
    A utility class for holding all the objective functions for each regressor algorithm in the search space.
        - You will see commonalities amongst the objective functions.  
        - Each objective function takes an MLautomator object and "parameter space" as arguments.  
        - During each pass through this objective function, Hyperopt selects a subset of the search space for 
          the appropriate algorithm in search_spaces.py.
        - Note that Hyperopt is also calling permutations of data transforms and feature selection as well.
        - Each objective function returns the mean cross-validated score, and the name of the algorithm.  
        - This gets anaylzed and packaged in the automator class itself.
    '''

    @staticmethod
    def objective01(automator, space):
        '''
        Objective function for XGBoost Regressor.
        '''
        algo = 'xgboost_regressor'
        X = automator.x_train
        Y = automator.y_train
        
        #Define the subset of dictionary keys that should get passed to the machine learning
        #algorithm.
        
        keys = get_keys(algo)
        subspace = {k: space[k] for k in set(space).intersection(keys)}
        
        #Extract the remaining keys that are pertinent to data preprocessing.
        
        model = XGBRegressor(n_jobs = -1, **subspace)     
        scaler = space.get('scaler')
        num_features = space.get('k_best')
        
        #Assemble a data pipeline with the extracted data preprocessing keys.
        pipeline = []
        pipeline = Pipeline([
            ('scaler', scaler),
            ('select_best', SelectKBest(k = num_features)),
            ('classifier', model),
        ])
        
        #perform cross validation and return the mean score.
        kfold = RepeatedKFold(n_splits = automator.num_cv_folds, n_repeats = automator.repeats)
        scores = -cross_val_score(pipeline, X, Y, cv = kfold, scoring = automator.score_metric, verbose = False).mean()   
        return scores, algo


    @staticmethod
    def objective02(automator, space):
        '''
        Objective function for SGD Regressor.
        '''
        algo = 'SGDRegressor'
        X = automator.x_train
        Y = automator.y_train
        #Define the subset of dictionary keys that should get passed to the machine learning
        #algorithm.
        
        keys = get_keys(algo)   
        subspace = {k: space[k] for k in set(space).intersection(keys)}
        
        #Extract the remaining keys that are pertinent to data preprocessing.
        model = SGDRegressor(**subspace)   
        scaler = space.get('scaler')
        num_features = space.get('k_best')
        
        #Assemble a data pipeline with the extracted data preprocessing keys.
        pipeline = []
        pipeline = Pipeline([
            ('scaler', scaler),
            ('select_best', SelectKBest(k = num_features)),
            ('classifier', model),
        ])
        
        #perform cross validation and return the mean score.
        kfold = RepeatedKFold(n_splits = automator.num_cv_folds, n_repeats = automator.repeats)
        scores = -cross_val_score(pipeline, X, Y, cv = kfold, scoring = automator.score_metric, verbose = False).mean()     
        return scores, algo


    @staticmethod
    def objective03(automator, space):
        '''
        Objective function for Random Forest Regressor.
        '''
        algo = 'RandomForestRegressor'
        X = automator.x_train
        Y = automator.y_train
        #Define the subset of dictionary keys that should get passed to the machine learning
        #algorithm.
        
        keys = get_keys(algo)  
        subspace = {k: space[k] for k in set(space).intersection(keys)}
        
        #Extract the remaining keys that are pertinent to data preprocessing.
        model = RandomForestRegressor(**subspace)   
        scaler = space.get('scaler')
        num_features = space.get('k_best')
        
        #Assemble a data pipeline with the extracted data preprocessing keys.
        pipeline = []
        pipeline = Pipeline([
            ('scaler', scaler),
            ('select_best', SelectKBest(k = num_features)),
            ('classifier', model),
        ])
        
        #perform two passes of 10-fold cross validation and return the mean score.
        kfold = RepeatedKFold(n_splits=10, n_repeats=1)
        scores = -cross_val_score(pipeline, X, Y, cv=kfold, scoring=automator.score_metric,verbose=False).mean()
        return scores, algo


    @staticmethod
    def objective04(automator, space):
        '''
        Objective function for Support Vector Machines. Note that this method uses a Bagged Classifier 
        as a wrapper for SVC.  Support Vector Machine run time scales by O(N^3).  Using bagged classifiers
        break up the dataset into smaller samples so that runtime is manageable.
        '''
        algo = 'SVR'
        X = automator.x_train
        Y = automator.y_train

        #Define the subset of dictionary keys that should get passed to the machine learning
        #algorithm.
        
        keys = get_keys(algo)    
        subspace = {k: space[k] for k in set(space).intersection(keys)}
 
        #Build a model with the parameters from our Hyperopt search space.

        n_estimators = space.get('n_estimators')
        model = BaggingRegressor(
            SVR(**subspace),
            max_samples = automator.num_samples // n_estimators,
            n_estimators = n_estimators,
            )   

        scaler = space.get('scaler')
        num_features = space.get('k_best')
        
        #Assemble a data pipeline with the extracted data preprocessing keys.
        pipeline = []
        pipeline = Pipeline([
            ('scaler', scaler),
            ('select_best', SelectKBest(k = num_features)),
            ('classifier', model),
        ])
        
        #perform cross validation and return the mean score.
        kfold = RepeatedKFold(n_splits = automator.num_cv_folds, n_repeats = automator.repeats)
        scores = -cross_val_score(pipeline, X, Y, cv=kfold, scoring = automator.score_metric, verbose=False, n_jobs=-1).mean()   
        return scores, algo 


    @staticmethod
    def objective05(automator, space):
        '''
        Objective function for K-Nearest Neighbors Voting Regressor.
        '''
        algo = 'KNeighborRegressor'
        X = automator.x_train
        Y = automator.y_train

        #Define the subset of dictionary keys that should get passed to the machine learning
        #algorithm.
        keys = get_keys(algo)    
        subspace = {k:space[k] for k in set(space).intersection(keys)}      

        #Build a model with the parameters from our Hyperopt search space.
        model = KNeighborsRegressor(n_jobs=-1, **subspace)
        scaler = space.get('scaler')
        num_features = space.get('k_best')
        
        #Assemble a data pipeline with the extracted data preprocessing keys.
        pipeline = []
        pipeline = Pipeline([
            ('scaler', scaler),
            ('select_best', SelectKBest(k = num_features)),
            ('classifier', model),
        ])
        
        #perform cross validation and return the mean score.
        kfold = RepeatedKFold(n_splits = automator.num_cv_folds, n_repeats = automator.repeats)
        scores = -cross_val_score(pipeline, X, Y, cv=kfold, scoring=automator.score_metric, verbose=False).mean()   
        return scores, algo   